@@ -14,6 +14,62 @@ var output = require("./widget_output");
 
 var MIME_TYPE = 'application/vnd.jupyter.widget-view+json';
 
+
+function polyfill_new_comm_buffers(manager, target_name, data, callbacks, metadata, comm_id, buffers) {
+    /**
+     * This polyfills services/kernel/comm/CommManager.new_comm to
+     * accept the buffers argument.
+     *
+     * argument comm_id is optional
+     */
+    return new Promise(function(resolve) {
+        requirejs(["services/kernels/comm"], function(comm) {
+            var comm = new comm.Comm(target_name, comm_id);
+            manager.register_comm(comm);
+            // inline Comm.open(), but with buffers
+            var content = {
+                comm_id : comm_id,
+                target_name : target_name,
+                data : data || {},
+            };
+            comm.kernel.send_shell_message("comm_open", content, callbacks, metadata, buffers);
+            resolve(comm);
+        });
+    });
+}
+
+function new_comm(manager, target_name, data, callbacks, metadata, comm_id, buffers) {
+    // Checks whether new_comm needs a polyfill, and calls the correct version
+    // Polyfill needed for notebook <5.1, in which the new_comm method does not support a buffers argument.
+    // See https://github.com/jupyter-widgets/ipywidgets/pull/1817
+    var need_polyfill = manager.new_comm.length < 6;
+    if (need_polyfill) {
+        return polyfill_new_comm_buffers.apply(null, arguments);
+    }
+    return manager.new_comm.apply(manager, Array.prototype.slice.call(arguments, 1));
+}
+
+/**
+ * Filter serialized widget state to remove any ID's already present in manager.
+ *
+ * @param {*} manager WidgetManager instance
+ * @param {*} state Serialized state to filter
+ *
+ * @returns {*} A copy of the state, with its 'state' attribute filtered
+ */
+function filter_existing_model_state(manager, state) {
+    var models = state.state;
+    models = Object.keys(models)
+        .filter(function(model_id) {
+            return !manager._models[model_id];
+        })
+        .reduce(function(res, model_id) {
+            res[model_id] = models[model_id];
+            return res;
+        }, {});
+    return _.extend({}, state, {state: models});
+}
+
 //--------------------------------------------------------------------
 // WidgetManager class
 //--------------------------------------------------------------------
@@ -28,24 +84,27 @@ var WidgetManager = function (comm_manager, notebook) {
     this.comm_manager = comm_manager;
 
     var widget_md = notebook.metadata.widgets
-    if (widget_md && widget_md['application/vnd.jupyter.widget-state+json']) {
-        this.set_state(notebook.metadata.widgets['application/vnd.jupyter.widget-state+json'])
-    }
 
-    // Register with the comm manager.
+    // Steps that needs to be done:
+    // 1. Register comm target
+    // 2. Get any widget state from the kernel and open comms with existing state
+    // 3. Check saved state for widgets, and restore any that would not overwrite
+    //    any live widgets.
+
+    // Register with the comm manager. (1)
     this.comm_manager.register_target(this.comm_target_name, _.bind(this.handle_comm_open,this));
 
-    // Attempt to reconstruct any live comms by requesting them from the back-end.
+    // Attempt to reconstruct any live comms by requesting them from the back-end (2).
     var that = this;
-    var backed_widgets_loaded = this._get_comm_info().then(function(comm_ids) {
+    this._get_comm_info().then(function(comm_ids) {
 
-        // Create comm class instances from comm ids.
+        // Create comm class instances from comm ids (2).
         var comm_promises = Object.keys(comm_ids).map(function(comm_id) {
             return that._create_comm(that.comm_target_name, comm_id);
         });
 
         // Send a state request message out for each widget comm and wait
-        // for the responses.
+        // for the responses (2).
         return Promise.all(comm_promises).then(function(comms) {
             return Promise.all(comms.map(function(comm) {
                 var update_promise = new Promise(function(resolve, reject) {
@@ -75,6 +134,14 @@ var WidgetManager = function (comm_manager, notebook) {
                     comm: widget_info.comm,
                 }, widget_info.msg.content.data.state);
             }));
+        }).then(function() {
+            // Now that we have mirrored any widgets from the kernel...
+            // Restore any widgets from saved state that are not live (3)
+            if (widget_md && widget_md['application/vnd.jupyter.widget-state+json']) {
+                var state = notebook.metadata.widgets['application/vnd.jupyter.widget-state+json'];
+                state = filter_existing_model_state(that, state);
+                return that.set_state(state);
+            }
         }).then(function() {
             // Rerender cells that have widget data
             that.notebook.get_cells().forEach(function(cell) {
@@ -210,12 +277,12 @@ WidgetManager.prototype.display_view = function(msg, view, options) {
 }
 
 
-WidgetManager.prototype._create_comm = function(comm_target_name, comm_id, data, metadata) {
+WidgetManager.prototype._create_comm = function(comm_target_name, comm_id, data, metadata, buffers) {
     var that = this;
     return this._get_connected_kernel().then(function(kernel) {
         if (data || metadata) {
-            return kernel.comm_manager.new_comm(comm_target_name, data,
-                                                that.callbacks(), metadata, comm_id);
+            return new_comm(kernel.comm_manager, comm_target_name, data,
+                            that.callbacks(), metadata, comm_id, buffers);
         } else {
             // Construct a comm that already is open on the kernel side. We
             // don't want to send an open message, which would supersede the
